#include "Arduino.h"
#include "typesDriver.h"


/*
Die Klasse Driver ist als Zustandsautomat realisiert. Die Zustände sind folgende:

straight,
right,
left

Sie entsprechen den zu Fahrenden Richtungen, um der Linie zu folgen.

Durch Aufruf der Methode driveStraight() fährt er ausschließlich geradeaus,
merkt sich aber weiterhin die Übergänge in neue Zustände.

Zwischen den Zuständen wird gewechselt, wenn die Linie verlassen oder wieder von den
Phototransistoren überdeckt wird.
*/

class Driver{

private:

  int dirL;   		// Direction-Pin Linker Stepper
  int stpL;   		// Step-Pin Linker Stepper
  int dirR;   		// Direction-Pin Rechter Stepper
  int stpR;   		// Step-Pin Rechter Stepper
  int stepDelay; 	// Delay zwischen steps. Reguliert die Dreh-Geschwindigkeit

  //  Phototransistoren 

  int photoL;	// Pin, an welchem der in Fahrtrichtung linke Phototransistor angeschlossen ist.
  int photoR; 	// Der rechte Phototransistor. In unserem Fall entspricht das dem mit dem roten Kabel.

  //  für Fahrfunktion

  int brightnessLeft;   // Variable für den Wert des Phototransitoren auf der linken Seite
  int brightnessRight; 	// Var. für den Wert des Phototransistoren auf der rechten Seite
  int lightThreshold; 	// Minimale Differenz, ab welcher die ausgegebenen Werte der Phototransistoren als unterschiedliche Untergründe interpretiert werden.
  int lightBlack;      	// Lichtintensität, ab welcher wir von Erkennung der schwarzen Linie ausgehen
  int photoCorrection; 	// Korrektur, die auf den rechten Phototransistor angewendet wird.

  DriverState currentState;  		// Speichert den aktuellen Zustand des Zustandsautomaten

public:

  /**
  Konstruktor des Drivers. Hier werden ihm alle relevanten Variablen übergeben.
  **/
  void setup(int newDirL, int newDirR, int newStpL, int newStpR, int newPhotoL, int newPhotoR){
  
  	// Setzen der Variablen
  	
    dirL = newDirL;   
    stpL = newStpL;   
    dirR = newDirR;  
    stpR = newStpR;
    photoL = newPhotoL;
    photoR = newPhotoR;

    stepDelay = 15;

    brightnessLeft = 0;     
    brightnessRight = 0;   	
    lightThreshold = 150;
    lightBlack = 600;  
    photoCorrection = 100;

	// Setzen des Initial-Zustands
	
    currentState = straight;

    // Stepper-Motoren initialisieren
    
    pinMode(dirL, OUTPUT);
    pinMode(stpL, OUTPUT);
    pinMode(dirR, OUTPUT);
    pinMode(stpR, OUTPUT);

    // Stepper-Motoren Richtung auf vorwärts stellen. 
    
    digitalWrite(dirL,LOW);
    digitalWrite(dirR,LOW);
  }
  
  void loop(){
  // wird nicht genutzt.
  }

  /**
  Methode um an der Linie entlang zu fahren.
  Der aktuelle Zustand bleibt auch zwischen Iterationen der Methode erhalten,
  daher kann der Roboter auch nachdem die Linie überhaupt nicht mehr gesehen wird zu ihr
  zurückfinden.
  **/
  void followLine(){
  
    // Auslesen der von den Phototransistoren gemessenen Helligkeiten. Jeweils Durchschnitt aus zwei Messungen.
    // Beachte: Unser Rechter Transistor gibt immer etwas weniger aus als der Linke, was wir durch die Variable photoCorrection korrigieren.
    brightnessLeft = (analogRead(photoL) + analogRead(photoL))/2;
    brightnessRight = (analogRead(photoR) + analogRead(photoR))/2 + photoCorrection;

	/*
     Überprüfe mit den Phototransitoren, wo die Linie ist 
     z.B. - wenn der linke Phototransistor wesentlich mehr Helligkeit misst als der Rechte,
     		können wir davon ausgehen, dass der Rechte über der Linie ist, der Linke aber nicht.
     		In dem Fall müssten wir nach Rechts gegenlenken.
     		Haben beide Transistoren die Linie verlassen, wird der currentState nicht geändert, sodass
     		er weiterhin in Richtung der Linie gegenlenkt.
    */
    if(brightnessLeft> brightnessRight + lightThreshold){  
      currentState = right;
    } 
    else if (brightnessRight > brightnessLeft + lightThreshold ){
      currentState = left;
    } 
    else if(brightnessRight < lightBlack && brightnessLeft < lightBlack){
      currentState = straight;
    }

    // Fahrbewegung entsprechend des aktuell gesetzten Zustands.
    switch(currentState){
    case right:  // nach rechts fahren
      digitalWrite(stpL, HIGH); 
      delay(1);               
      digitalWrite(stpL, LOW);  
      delay(stepDelay);
      break;
    case left:   // nach links fahren
      digitalWrite(stpR, HIGH);   
      delay(1);               
      digitalWrite(stpR, LOW);  
      delay(stepDelay);
      break;
    case straight:   // geradeaus fahren
      digitalWrite(stpL, HIGH);
      digitalWrite(stpR, HIGH);   
      delay(1);
      digitalWrite(stpL, LOW);                 
      digitalWrite(stpR, LOW);  
      delay(stepDelay);
      break; 
    }
  }
  
  /**
  driveStraight() wird aufgerufen, wenn nur geradeaus gefahren werden soll.
  Die Phototransistoren werden trotzdem ausgelesen und die Zustände entsprechend geändert,
  um dem Roboter zu ermöglichen, wieder zur Linie zurück zu finden.
  **/
  void driveStraight(){
    
    // Die Zustände werden genauso gesetzt, wie bei der Methode follow().
    brightnessLeft = (analogRead(photoL) + analogRead(photoL))/2;
    brightnessRight = (analogRead(photoR) + analogRead(photoR))/2 + photoCorrection;

    if(brightnessLeft> brightnessRight + lightThreshold){  
      currentState = right;
    } 
    else if (brightnessRight > brightnessLeft + lightThreshold ){
      currentState = left;
    } 
    else if(brightnessRight < lightBlack && brightnessLeft < lightBlack){
      currentState = straight;
    }
    
    // Führt bei beiden Steppern einen Schritt in die selbe Richtung aus. Fährt also einen Schritt vorwärts.
    digitalWrite(stpL, HIGH);
    digitalWrite(stpR, HIGH);   
    delay(1);
    digitalWrite(stpL, LOW);                 
    digitalWrite(stpR, LOW);  
    delay(stepDelay);
  }

};

