  /** 
  Hauptklasse für den Giessroboter Nessie.
  
  Nessie folgt einer auf einem glatten weißen Untergrund angebrachten schwarzen Linie,
  erkennt rechts dieser Linie aufgestellte Shotgläser mit Hilfe eines Sonars und befüllt diese
  über einen Arm mit einer Flüssigkeit.
  
  Realisiert als Zustandsautomat mit den Zuständen aus der Klasse types.h (Siehe dort für 
  mehr Informationen zu den Zuständen).
  **/

  #pragma once //damit Klassen nur einmalig eingefügt werden
  #include <Servo.h>    // Import der Servo-Libraries 
  #include "types.h"    //Zustandstypen
  #include "Driver.h" 	//Fahrkompenente
  #include "Filler.h"	//regelt Servobewegungen, d.h. Arm und Ventil
  
    // Pin-Belegung auf dem Arduino
  
        // Start-Knopf
    
        const int startButton = A3;
  
        //  Stepper - Motoren
        /* 
           Zusatz-Infos zu den Steppern:
       
            Pololu 1 für den Linken Stepper befindet sich direkt neben dem Arduino, 
            Pololu 2 für den Rechten Stepper ist der Andere.
            Den Linken Stepper so anbringen, dass das schwarze Kabel direkt neben dem 
            Kondensator ist, den Rechten so, dass das blaue Kabel direkt daneben ist.
        */
    
        const int dirL = 2; // Linker Stepper
        const int stpL = 3;   
        const int dirR = 7; // Rechter Stepper
        const int stpR = 9;   
        const double stepLength = 1.414/2; // Länge eines Halb-Schrittes mit den 90mm Rädern.
   
        // Sonar 
    
        const int inPinSonar = A4;
        const int pingPinSonar = 12;
    
        // Arm-Servos für Drehbewegung + Heben
    
        const int rotor = 6;     //mittlerer Anschluss
        const int lifter = 4;    //innerer Anschluss
    
        // Servo mit Ventil-Funktion - 0 ist weitestmöglich Richtung geschlossen, 180 ist weitestmöglich geöffnet.
    
        const int vent = 5;  //äußerer Anschluss

        //  LED
    
        const int resetLed = 13; // LED auf dem Arduino, soll Leuchten, wenn der Arduino abstürzt und sich resetted
    
        //  Phototransistoren 
    
        const int photoL = A6;
        const int photoR = A7; // Das blaue Kabel
    
        // Kontakt für die Elektrode zur Erkennung eines vollen Glases
        // (auf dem Breadboard angebracht: Erdung G56 (schwarz), Stromversorgung G55 (rot))
    
        const int electrodePin = A5;
    
    //  Konstanten

  	const int armLength = 229; //  Armlänge in mm
    	const int distSonar = 315; // Entfernung des Mittelpunkts des Sonars von der Dreh-Achse des Arms in mm
    	const double radiusGlass = 22.5; // Radius des Glases in mm
    	const double alpha=3.15/6; // Kegelöffnung Sonar in rad

    //  Variablen  

        // für Glaserkennung
        double glassDistance = 0;
        // Speichervariablen
        double slantHeight = 0; // Mantellinie
        double mitteSonar = 0;  // berechneter Kegelradius
        double steps = 0;

    // Objekte

    Filler filler; // Füll-Bewegungen
    Driver driver; //Fahr-Bewegungen
    
    // Zustand
    
    State currentState = standby;
    
    /**
    Setup-Methode, in der nur Objekte initialisiert werden.
    **/
    void setup(){
      
      // Lässt die LED auf dem Arduino für 300ms aufleuchten, wenn der Arduino neu startet.
      // Erlaubt uns zu erkennen, wenn er abgestürzt ist.
      pinMode(resetLed,OUTPUT);
      digitalWrite(resetLed,HIGH);
      delay(300);
      digitalWrite(resetLed,LOW);    
      
      //Initialisierungen
      driver.setup(dirL,dirR,stpL,stpR,photoL,photoR);
      filler.setup(rotor,lifter,vent,electrodePin);

    }

    /**
    Hauptmethode, welche den Ablauf steuert, d.h. Linie folgen, Glas finden, Arm bewegen,
    Füllen und wieder von vorn beginnen.
    **/
    void loop(){
        /*
        Merke: Jedes Objekt, das auf weniger als die Armlänge des Roboters
        von ihm entfernt zu seiner Rechten aufgestellt ist, wird als Glas erkannt.
        Steht das gLas weiter weg, als 18cm wird es nicht erkannt.
        */
    
        switch(currentState){
          case standby:
            // Hier wird abgefragt, ob der Startknopf gedrückt worden ist.
            if(analogRead(startButton)==1023) currentState = follow;
            break;
          case follow:
            //folge der Linie
            for(int i = 0; i < 5; i++){
                driver.followLine();
            }
            // Frage das Sonar alle 5 Schritte, ob sich ein Glas in der Nähe befindet,  
            // wenn ja, gehen wir in den Zustand found
            currentState = detectGlass();
            break;
          case found:
            //finde die Mitte, slantHeight wurde in detectGlass() angepasst
            mitteSonar = calcCenter(slantHeight) + radiusGlass;
        	// Fahre zur Mitte
            steps = convertToSteps(mitteSonar);
            for(int i = 0; i < steps ; i++){
                 driver.driveStraight();
                 delay(10);
             }
        	 // Messe Glasdistanz
             glassDistance = distanceSonar(); 
             //Wechsel in den nächsten Zustand
             currentState = fill;
             break;
          case fill:
            // Passe die Distanz so an, dass sie bis zur Mitte des Glases geht
            glassDistance = glassDistance + radiusGlass; 
            // Berechne die zu fahrende Restdistanz
            double h = calcH(); 
            // Korrigiere den Wert hinsichtlich Messfehler
            // Vergrößere sie abhängig davon, wie nah das Glas dran steht
            h = map(h, 0, 316, 15, 437); 
            // Fahre die Restdistanz
            double steps = convertToSteps(h);
            for(int i=0; i< steps; i++){
              driver.driveStraight();
            }
        	// Bewege den Arm, sodass der Schlauch das Glas trifft
            filler.moveArm(glassDistance); 
      
            //fülle Glas 
            filler.fillGlass();
        
            //alles wieder auf Anfang
            filler.resetArm();
            glassDistance =0;
            currentState = follow;
     
            break; 
        }     
    }
    
    /**
    Wechselt den Zustand von follow zu found, wenn ein Glas entdeckt wird.
    **/
    State detectGlass(){
    double distance = distanceSonar();
    if(distance > 0){
     // Speichere die Distanz zwischen um sie später als Mantellinie weiter zu verwenden
      slantHeight = distance; 
      return found;
    }
    else return follow;
    }
    
    /**
    Als Eingabe bekommen wir die Mantellinie des Sichtkegels des Sonars. 
    Ausgehend davon berechnet die Methode jene Distanz sodass das Sonar direkt vor dem Glas steht.
    (Das ist der Kegelradius).
    **/
    double calcCenter(double s){ 
    //Korrekturwert 5*stepLength, da nur alle 5 Schritte nach einem Glas geschaut wird
    return (s + 5 * stepLength) * tan(alpha/2) / (sqrt(1+tan(alpha/2)*tan(alpha/2)));
    }

    /**
    Methode, welche mit dem Sonar die Distanz zu etwaigen Objekten überprüft und uns in mm zurück gibt.
    Gibt es kein Objekt in direkter Nähe (d.h. innerhalb von 18cm), wird 0 zurück gegeben. 
    **/
    double distanceSonar(){

    // Initialisiere Variablen für die Distanz 
    // duration sind ms und mm sind millimeter
    double duration = 0;
    double mm;
    
    // Um einen Durchschnittswert der Messungen zu bilden, messen wir 10 mal
    for(int i = 0; i < 10; i++){
      // Um ein einen Sonar-Ping auszusenden, sendet man dem Trigger-Pin einen HIGH-Puls von 2 oder mehr Mikrosekunden
      // Um einen sauberen HIGH-Puls zu gewährleisten, senden wir vorher einen kurzen LOW-Puls.
      pinMode(pingPinSonar, OUTPUT);
      digitalWrite(pingPinSonar, LOW);
      delayMicroseconds(2);
      digitalWrite(pingPinSonar, HIGH);
      delayMicroseconds(10);
      digitalWrite(pingPinSonar, LOW);
      
      // Der Echo-Pin wird dann genutzt, um das Echo des Pings zu empfangen.
      // Dabei entspricht die Länge des HIGH-Pulses, der vom Echo-Pin ausgeht
      // der Dauer zwischen Aussenden des Pings und des ersten Empfangs des Echos.
      pinMode(inPinSonar, INPUT);
      duration = duration + pulseIn(inPinSonar, HIGH); // Duration wird zur Ermittlung des Mittelwerts in jedem Durchgang aufaddiert.
      // Mit neueren Arduino IDEs kann hier auch pulseIn(inPinSonar,HIGH,20000) genutzt werden, 
      // wobei die 20000 einem Timeout nach 200000 Mikrosekunden entspricht. Alte Versionen haben hier einen Bug, daher ist Vorsicht geboten.
      delay(25);
     }
    // in mm umrechnen 
    mm = microsecondsToMillimeters(duration/10); 

    // Maximale Entfernung vom Sonar, für die wir das Glas noch erreichen können ist 180mm,
    // daher werden Objekte in Entfernungen > 180mm ignoriert.
    if(mm > 180) mm = 0;

    return mm;
    }

    /**
    Restlänge, die der Roboter zurücklegen muss, wenn das Sonar vor dem Glas steht. 
    **/
    double calcH(){
    double square = sqrt(sq((double) armLength) - sq((double)glassDistance)); 
    return distSonar - square; 
    }

    /**
    Methode Um MicroSekunden in mm umzurechnen. 
    Schallgeschwindigkeit liegt bei 340 m/s oder 29 mikrosekunden/cm. 
    Der Ping geht den Weg zum Objekt und zurück, also rechnen wir durch 2.
    **/
    double microsecondsToMillimeters(double microseconds){
     return (microseconds / 2.9 / 2);
    }
   
    /**
      Rechnet mm in Stepperschritte um.
    **/
    double convertToSteps(double mm){
      return mm/stepLength;
    }




